package NoRFC::Server; # $Id: Server.pm 198 2021-03-11 15:07:38Z minus $
use strict;
use warnings FATAL => 'all';
use utf8;

=encoding utf8

=head1 NAME

NoRFC::Server - NoRFC web-server class

=head1 VERSION

Version 1.02

=head1 SYNOPSIS

    use NoRFC::Server;

=head1 DESCRIPTION

This module provides NoRFC web-server functionality

=head2 reload

The reload hook

=head2 startup

Mojo application startup method

=head1 SEE ALSO

L<Mojolicious>, L<NoRFC>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2022 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use Mojo::Base 'Mojolicious';

use Mojo::File qw/path/;
use Mojo::Util qw/sha1_sum secure_compare/;

use CTK::Util qw/preparedir sharedir sharedstatedir/;
use CTK::ConfGenUtil;

use vars qw/ $VERSION /;
$VERSION = "1.02";

has 'documentroot';

# Mojo Routes (paths)
sub startup {
    my $self = shift; # app
    my $ctk = $self->can('ctk') ? $self->ctk() : $self->{ctk};

    # Set password
    my $username = value($ctk->conf("username") // "");
    my $secret = value($ctk->conf("password") // "");
    $self->secrets([$secret]) if length($secret);

    # Logging
    $self->log->level("warn")->path($ctk->logfile()) unless $ctk->debugmode();

    #$self->log->debug("Startup!! =$$"); # $self->ctk->logdir()

    # Switch to installable home directory
    $self->home(Mojo::Home->new($ctk->datadir()));

    # Get DocumentRoot and replace as public-path
    my $documentroot = value($ctk->conf("documentroot")) || path(sharedir(), $ctk->prefix)->to_string();
    $self->documentroot($documentroot);
    $self->static->paths()->[0] = $documentroot; #unshift @{$static->paths}, '/home/sri/themes/blue/public';

    # Hooks
    $self->hook(before_dispatch => sub {
        my $c = shift;

        # Set Server header
        $c->res->headers->server(sprintf("%s/%s", NoRFC::PROJECTNAME(), NoRFC->VERSION));

        # Parse URL path
        my %norfcnum = $c->app->ctk->parse_norfcnum($c->req->url->path);
        my $need_auth = ($norfcnum{status} && $norfcnum{status} eq "P") ? 1 : 0;
        my $forbidden = ($norfcnum{status} && $norfcnum{status} eq "Z") ? 1 : 0;
        #$c->app->log->info(" !!!! " . $need_auth); # $self->ctk->logdir()

        # Authentication
        if ($need_auth && length($username)) {
            my $req_uri = $c->req->url->to_abs();
            my $ui_username = $req_uri->username() // "";
            my $ui_secret = sha1_sum($req_uri->password() // time());

            # Check username and password
            return 1 if length($secret)
                and secure_compare($username, $ui_username)
                and secure_compare($secret, $ui_secret);

            # Require authentication
            $c->res->headers->www_authenticate('Basic realm="NoRFC Strict Zone"');
            $c->render(json => {
                status => 0,
                message => "Authentication required!",
            }, status => 401);
        } elsif ($forbidden) {
            $c->render(json => {
                status => 0,
                message => "Forbidden!",
            }, status => 402);
        }
    });

    # Routes
    $self->routes->get('/')->to('alpha#root');
    $self->routes->get('/:rfcnum' => [rfcnum => NoRFC::RFCNUM_PATTERN()])->to('alpha#norfc'); # Draft-R21047017
    $self->routes->get('/env')->to('alpha#env') if $ctk->debugmode();
    $self->routes->get('/info')->to('alpha#info');
    $self->routes->post('/search')->to('alpha#search');

    # Favicon
    delete $self->static->extra->{'favicon.ico'};

    return 1;
}

# Reload hook
sub reload {
    my $self = shift;
    $self->log->warn("Request for reload $$"); # $self->ctk->logdir()
    return 1; # 1 - ok; 0 - error :(
}

1;

# Controller
package NoRFC::Server::Alpha;
use utf8;
use Mojo::Base 'Mojolicious::Controller';

use Mojo::File qw/path/;
use Mojo::JSON qw/from_json/;
use Mojo::Util qw/b64_decode/;

use Try::Tiny;

use CTK::Util qw/sharedir sharedstatedir trim dformat/;
use CTK::ConfGenUtil;

use constant {
        CONTENT_TYPE_JSON   => "application/json",
        CONTENT_TYPE_TXT    => "text/plain; charset=utf-8",
        ROW_NUM             => 500,
    };

sub root {
    my $self = shift;
    return $self->reply->static('index.html');
}
sub env {
    my $self = shift;
    $self->res->headers->content_type(CONTENT_TYPE_TXT);
    return $self->render(text => $self->dumper(\%ENV));
}
sub info {
    my $self = shift; # Mojolicious::Routes
    my $ctk = $self->app->ctk();
    my $store = $ctk->store;
    my $cnt = $store->count || 0;

    #$self->app->log->debug("Catched!");

    return $self->render(json => {
            name => "info",
            status => $store->status,
            message => $store->status ? "OK" : $store->error || sprintf("Store (%s): Unknown error", $store->dsn),
            total => $cnt,
            version => NoRFC->VERSION,
            categories => [(@{(NoRFC::RFC_CATEGORIES())}, @{(NoRFC::EXTRA_CATEGORIES())})],
            $ctk->debugmode() ? (
                datadir => $ctk->datadir,
                homedir => $self->app->home(),
                documentroot => $self->app->documentroot,
                paths => $self->app->static->paths(),
            ) : (),
        });
}
sub search {
    my $self = shift;
    my $ctk = $self->app->ctk();
    my $store = $ctk->store;
    my $raw = $self->req->body // '';

    # Incorrect data from STDIN (as POST data)
    return $self->render(json => {
            status => 0,
            message => sprintf("Incorrect POST data"),
        }, status => 400) unless length($raw);

    # Deserialize data
    my $data;
    my $err;
    try {
        $data = from_json($raw);
    } catch {
        $err = $_;
    };
    return $self->render(json => {
            status => 0,
            message => sprintf("Incorrect JSON: %s", $err),
        }, status => 400) unless $data && is_hash($data);

    # Get data for search
    my $in_norfc = uc(trim($data->{norfc} || ''));
    my $in_category = trim($data->{category} || '');
    my $in_title = trim($data->{title} || '');

    # Store check
    return $self->render(json => {
            status => 0,
            message => $store->error || sprintf("Store (%s): Unknown error", $store->dsn),
        }, status => 500) unless $store->status;

    # Get
    my $total = 0;
    my @arr = ();
    my @found = ();
    if ($in_norfc) {
        $in_norfc =~ s/DRAFT/Draft/i;
        $in_norfc =~ s/NORFC/RFC/i;
        my %info = $store->get($in_norfc);
        #my $cat = $info{category} || '';
        push @arr, {
            n           => ++$total,
            id          => $info{id} || 0,
            norfc       => $info{num} || '',
            published   => $info{pub} || 0,
            category    => $info{category} || '',
            title       => $info{title} || '',
        } if $info{id}; # && (!$in_category || ($in_category && lc($cat) eq $in_category));
    } else {
        @found = $store->search(
            row_count   => ROW_NUM,
            category    => $in_category,
            title       => $in_title,
        );
    }
    return $self->render(json => {
            status => 0,
            message => $store->error || sprintf("Store (%s): Unknown error", $store->dsn),
        }, status => 500) unless $store->status;

    foreach my $row (@found) {
        push @arr, {
                n           => ++$total,
                id          => $row->[0] || 0,
                norfc       => $row->[1] || '',
                published   => $row->[2] || 0,
                category    => $row->[3] || '',
                title       => $row->[4] || '',
        };
    }

    # Output
    return $self->render(json => {
            status => 1,
            message => "",
            name => "search",
            data => [@arr],
            in => {
                norfc       => $in_norfc,
                category    => $in_category,
                title       => $in_title,
            },
        });
}
sub norfc {
    my $self = shift;
    my $ctk = $self->app->ctk();
    my $rfcnum = $self->param("rfcnum");
    my $store = $ctk->store;

    # Check number
    return $self->render(json => {
            status => 0,
            message => "Incorrect document number",
        }, status => 400) unless $ctk->is_valid($rfcnum, NoRFC::QUIET_MODE());

    # Check store instance
    return $self->render(json => {
            status => 0,
            message => $store->error || sprintf("Store (%s): Unknown error", $store->dsn),
        }, status => 500) unless $store->status;


    # Get document info from DB
    my %info = $store->get($rfcnum);
    return $self->render(json => {
            status => 0,
            message => $store->error || sprintf("Store (%s): Unknown error", $store->dsn),
        }, status => 500) unless $store->status;
    return $self->render(json => {
            status => 0,
            message => "Document not found",
        }, status => 404) unless $info{id};

    # Create document
    my $doc = dformat(NoRFC::NORFC_TEMPLATE(), {
        NORFC_RFCNUM    => $rfcnum,
        NORFC_PUB       => NoRFC::_fdate($info{pub}, 1),
        NORFC_CATEGORY  => $info{category} || NoRFC::DEFAULT_CATEGORY(),
        NORFC_AUTHOR    => $info{author} || NoRFC::DEFAULT_AUTHOR(),
        NORFC_TITLE     => $info{title} || NoRFC::DEFAULT_TITLE(),
        NORFC_CONTENT   => $info{content} || NoRFC::DEFAULT_CONTENT(),
    });
    Encode::_utf8_on($doc);

    $self->res->headers->content_type(CONTENT_TYPE_TXT);
    return $self->render(text => $doc);
}

1;

__END__
